import { READ_STATE } from "@src/const";
import type { CachedMetadata, TFile } from "obsidian";
import { getContext, setContext } from "svelte";
import { getSettingsContext } from "./settings.svelte";
import type BookTrackerPlugin from "@src/main";
import { createDateFilter, type DateFilterStore } from "./date-filter.svelte";

export type FileMetadata = {
	file: TFile;
	frontmatter: Record<string, any>;
};

export type FileProperty = {
	file: TFile;
	value: any;
};

interface MetadataStore extends DateFilterStore {
	get metadata(): FileMetadata[];

	destroy(): void;
}
export function createMetadata(plugin: BookTrackerPlugin): MetadataStore {
	const settings = getSettingsContext().settings;

	const initialMetadata: FileMetadata[] = [];

	for (const file of plugin.app.vault.getMarkdownFiles()) {
		const frontmatter =
			plugin.app.metadataCache.getFileCache(file)?.frontmatter ?? {};

		if (frontmatter[settings.statusProperty] !== READ_STATE) {
			continue;
		}

		initialMetadata.push({ file, frontmatter });
	}

	let metadata = $state(initialMetadata);

	function onChanged(file: TFile, _data: string, cache: CachedMetadata) {
		metadata = metadata.map((f) => {
			if (f.file.path === file.path) {
				return {
					...f,
					frontmatter: cache.frontmatter ?? {},
				};
			}
			return f;
		});
	}

	plugin.registerEvent(plugin.app.metadataCache.on("changed", onChanged));

	function onDeleted(file: TFile) {
		metadata = metadata.filter((f) => f.file.path !== file.path);
	}
	plugin.registerEvent(plugin.app.metadataCache.on("deleted", onDeleted));

	const dateFilter = createDateFilter(
		() => metadata,
		(f) => {
			// @ts-expect-error Moment is provided by Obsidian
			return moment(f.frontmatter[settings.endDateProperty]);
		}
	);

	return {
		get metadata() {
			return dateFilter.filteredData;
		},
		get filterYear() {
			return dateFilter.filterYear;
		},
		set filterYear(value) {
			dateFilter.filterYear = value;
		},
		get filterMonth() {
			return dateFilter.filterMonth;
		},
		set filterMonth(value) {
			dateFilter.filterMonth = value;
		},
		get filterYears() {
			return dateFilter.filterYears;
		},
		get filterMonths() {
			return dateFilter.filterMonths;
		},
		destroy() {
			plugin.app.metadataCache.off("changed", onChanged);
			plugin.app.metadataCache.off("deleted", onDeleted);
		},
	};
}

const METADATA_KEY = Symbol("metadata");

export function setMetadataContext(state: MetadataStore) {
	setContext(METADATA_KEY, state);
}

export function getMetadataContext(): MetadataStore {
	return getContext(METADATA_KEY) as MetadataStore;
}

function notEmpty(value: any): boolean {
	return value !== undefined && value !== null && value !== "" && value !== 0;
}

interface PropertyStore {
	get propertyData(): FileProperty[];
}

export function createPropertyStore(
	property: string,
	filter: (value: any) => boolean = notEmpty
): PropertyStore {
	const store = getMetadataContext();

	const propertyData = $derived(
		store.metadata
			.map((f) => ({ ...f, value: f.frontmatter[property] }))
			.filter((f) => (filter ? filter(f.value) : true))
	);

	return {
		get propertyData() {
			return propertyData;
		},
	};
}
